"""
Composable functions for extracting data using
`lxml <http://lxml.de/>`_.
"""

from __future__ import absolute_import, unicode_literals, print_function
import wex.py2compat ; assert wex.py2compat  # flake8: noqa
import logging
from itertools import islice, chain
from copy import deepcopy
from operator import methodcaller, itemgetter
from six import string_types
from six.moves import map, reduce
from six.moves.urllib_parse import urljoin, quote, unquote
from lxml.etree import (XPath,
                        _ElementTree,
                        _Element,
                        Element,
                        FunctionNamespace)
from lxml.cssselect import CSSSelector
from lxml.html import XHTML_NAMESPACE, HTMLParser

from .composed import composable, Composable
from .cache import cached
from .iterable import _do_not_iter_append, filter_if_iter
from .htmlstream import HTMLStream
from .ncr import replace_invalid_ncr
from .url import URL, public_suffix


NEWLINE = u'\n'
EMPTY = u''
SPACE = u' '


# we do not want to flatten etree elements
_do_not_iter_append(_Element)

UNPARSEABLE = Element('unparseable')

base_href = XPath('//base[@href]/@href | //x:base[@href]/@href',
                  namespaces={'x': XHTML_NAMESPACE})

default_namespaces = {'re': 'http://exslt.org/regular-expressions'}

# see http://lxml.de/extensions.html#the-functionnamespace
function_namespace = FunctionNamespace(None)

_html_text_nodes = XPath(
    'descendant-or-self::node()' +
    '[not(local-name()) or not(text())]' +
    '[not(ancestor::script or ancestor::style)]'
)


def _wex_html_text(context, arg=None):
    if arg is None:
        arg = [context.context_node]
    html_text = []
    for node in chain.from_iterable(map(_html_text_nodes, arg)):
        tag = getattr(node, 'tag', None)
        if tag is None:
            html_text.append(node)
        elif tag == 'br':
            html_text.append(NEWLINE)
        else:
            html_text.append(EMPTY)
    return EMPTY.join(html_text)

function_namespace['wex-html-text'] = _wex_html_text



@composable
@cached
def parse(src):
    """ Returns an element tree create by `LXML <http://lxml.de/>`_.
       :param src: A readable object such as a :class:`wex.response.Response`.
    """

    if not hasattr(src, 'read'):
        return src

    etree = _ElementTree()
    try:
        stream = HTMLStream(src)
        # Sometimes we get URLs containing characters that aren't
        # acceptable to lxml (e.g. "http:/foo.com/bar?this=array[]").
        # When this happens lxml will quote the whole URL.
        # We don't want to have to check for this so we just always
        # quote it here and then unquote it in the `base_url` function.
        quoted_base_url = quote(src.url) if src.url else src.url
        while True:
            try:
                fp = replace_invalid_ncr(stream)
                # fp is a Unicode stream
                # The lxml FAQ tells us that it is inefficient to do this
                # http://lxml.de/FAQ.html#can-lxml-parse-from-file-objects-opened-in-unicode-text-mode
                # but actually it seems just fine as long as you tell the parser to use 'utf-8'!?
                parser = HTMLParser(encoding='utf-8')
                etree.parse(fp, parser=parser, base_url=quoted_base_url)
                break
            except UnicodeDecodeError as exc:
                stream.next_encoding()
    except IOError as exc:
        logger = logging.getLogger(__name__)
        logger.warning("IOError parsing %s (%s)", src.url, exc)

    root = etree.getroot()
    if root is None:
        etree._setroot(UNPARSEABLE)

    return etree


@cached
def get_base_url_from_root(root):
    if root.base_url:
        # see :func:`.parse` for why we need to unquote
        base_url = unquote(root.base_url)
    else:
        base_url = root.base_url
    return reduce(urljoin, base_href(root)[:1], base_url)


def get_base_url(elem_or_tree):
    if hasattr(elem_or_tree, 'getroottree'):
        tree = elem_or_tree.getroottree()
    else:
        # if it doesn't have getroottree() we presume it's a tree!
        tree = elem_or_tree
    return get_base_url_from_root(tree.getroot())


class map_if_list(Composable):

    def __init__(self, func):
        self.func = func

    def __repr__(self):
        return '%s(%r)' % (self.__class__, self.func)

    def __compose__(self):
        return (self,)

    def __call__(self, *args, **kwargs):
        if args and isinstance(args[0], list):
            return [res for res in map(self.func, *args, **kwargs)]
        return self.func(*args, **kwargs)


def css(expression):
    """ Returns a :func:`composable <wex.composed.composable>` callable that
        will select elements defined by a
        `CSS selector <http://en.wikipedia.org/wiki/Cascading_Style_Sheets#Selector>`_
        expression.

        :param expression: The CSS selector expression.

        The callable returned accepts a :class:`wex.response.Response`, a
        list of elements or an individual element as an argument.
    """
    return parse | map_if_list(CSSSelector(expression))


def xpath(expression, namespaces=default_namespaces):
    """ Returns :func:`composable <wex.composed.composable>` callable that will
        select elements defined by an
        `XPath <http://en.wikipedia.org/wiki/XPath>`_ expression.

        :param expression: The XPath expression.
        :param namespaces: The namespaces.

        The callable returned accepts a :class:`wex.response.Response`, a
        list of elements or an individual element as an argument.

        For example:

        .. code-block:: pycon

            >>> from lxml.html import fromstring
            >>> tree = fromstring('<h1>Hello</h1>')
            >>> selector = xpath('//h1')

    """
    return parse | map_if_list(XPath(expression, namespaces=namespaces))


def attrib(name, default=None):
    getter = methodcaller('get', name, default)
    return map_if_list(getter)


def base_url_pair_getter(get_url):
    """ Returns a function for gettting a tuple of `(base_url, url)` when
        called with an etree `Element` or `ElementTree`.

        In the returned pair `base_url` is the value returned from
        `:func:get_base_url` on the etree `Element` or `ElementTree`.
        There second value is the value returned by calling the `get_url`
        on the same the same etree `Element` or `ElementTree`, joined to
        the `base_url` using `urljoin`.  This allows `get_url` to return
        a relative URL.
    """
    @composable
    def get_base_url_pair(elem_or_tree):
        base_url = get_base_url(elem_or_tree)
        url = get_url(elem_or_tree)
        if url:
            url = URL(urljoin(base_url, url.strip()))
        return (URL(base_url), url)
    return get_base_url_pair


def same_domain(url_pair):
    """ Return second url of pair if both are from same domain. """

    if not all(url_pair):
        return None

    base_url, url = map(URL, islice(url_pair, 2))
    if base_url.parsed[:2] == url.parsed[:2]:
        return url

    return None


def same_suffix(url_pair):
    """ Return second url of pair if both have the same public suffix. """

    if not all(url_pair):
        return None

    base_url, url = map(URL, islice(url_pair, 2))

    if url.parsed.hostname is None:
        return None

    base_suffix = public_suffix(base_url)
    dot_suffix = '.' + base_suffix
    dot_hostname = '.' + url.parsed.hostname
    if dot_hostname.endswith(dot_suffix):
        return url



src_base_url_pair = base_url_pair_getter(methodcaller('get', 'src'))
href_base_url_pair = base_url_pair_getter(methodcaller('get', 'href'))

# helpers that operate on exactly one element
src_url_1 = src_base_url_pair | itemgetter(1)
href_url_1 = href_base_url_pair | same_domain
href_url_same_suffix_1 = href_base_url_pair | same_suffix
href_any_url_1 = href_base_url_pair | itemgetter(1)


#: A :class:`wex.composed.ComposedFunction` that returns the absolute
#: URL from an ``href`` attribute as long as it is from the same domain
#: as the base URl of the response.
href_url = map_if_list(href_url_1) | filter_if_iter(bool)

#: A :class:`wex.composed.ComposedFunction` that returns the absolute
#: URL from an ``href`` attribute as long as it is from the same
#: `public suffix <https://publicsuffix.org/>`_
#: as the base URl of the response.
href_url_same_suffix = (map_if_list(href_url_same_suffix_1) |
                        filter_if_iter(bool))

#: A :class:`wex.composed.ComposedFunction` that returns the absolute
#: URL from an ``href`` attribute.
href_any_url = map_if_list(href_any_url_1) | filter_if_iter(bool)


#: A :class:`wex.composed.ComposedFunction` that returns the absolute
#: URL from an ``src`` attribute.
src_url = map_if_list(src_url_1) | filter_if_iter(bool)


def itertext(*tags, **kw):
    """ Return a function that will return an iterator for text.  """
    with_tail = kw.pop('with_tail', True)
    if kw:
        raise ValueError('unexpected keyword arguments %s' % kw.keys())

    @composable
    def _itertext(src):
        if hasattr(src, 'itertext'):
            return src.itertext(*tags, with_tail=with_tail)
        elif hasattr(src, '__iter__') and not isinstance(src, string_types):
            text_nodes = (_itertext(i) for i in src)
            return chain.from_iterable(text_nodes)
        raise ValueError("%r is not iterable" % src)
    return _itertext


def drop_tree(*selectors):
    """ Return a function that will remove trees selected by `selectors`. """

    @map_if_list
    def tree_dropper(src):
        copied = None
        for selector in selectors:
            selected = selector(copied if copied is not None else src)
            if selected and copied is None:
                copied = deepcopy(src)
                selected = selector(copied)
            for selection in selected:
                selection.drop_tree()
        return copied if copied is not None else src
    return tree_dropper


@map_if_list
def normalize_space(obj):
    """ Normalize space according to standard Python rules.

    The definition of what is space used in XPath's
    `normalize-space <http://www.w3.org/TR/xpath/#function-normalize-space>`_
    is a small subset of the characters defined as space in the
    `unicode <https://en.wikipedia.org/wiki/Whitespace_character#Unicode>`_
    rules that Python uses.
    """
    if hasattr(obj, 'split'):
        obj_as_text = obj
    else:
        obj_as_text = text_content(obj)
    return SPACE.join(obj_as_text.split())


def list2set(obj):
    if isinstance(obj, list):
        return set(obj)
    return obj


#: Return text content from an object (typically node-set) excluding from
#: content from within `<script>` or `<style>` elements.
text_content = xpath('wex-html-text(.)')


#: Alias for `normalize-space | list2set`
text = normalize_space | list2set



""" Content Decoders HTTP

Ideally requests/urllib3 would do the decoding for us, but currently we
cannot rely on that because the `.raw.read()` can return b'' simply
because the decompressor doesn't have enough to work on.

See https://github.com/shazow/urllib3/issues/709

This is most of the work-around.  The other part is a small change
in wex/http.py
"""

from __future__ import unicode_literals
import io
import zlib

CHUNK = 1024 * 8


class ZDecoder(io.RawIOBase):
    """ Base class for HTTP content decoders based on zlib """

    def __init__(self, fp, z=None):
        self.fp = fp
        self.z = z
        self.flushed = None

    def readinto(self, buf):

        if self.z is None:
            self.z = zlib.decompressobj()
            retry = True
        else:
            retry = False

        n = 0
        max_length = len(buf)

        while max_length > 0:

            if self.flushed is None:

                chunk = self.fp.read(CHUNK)
                compressed = (self.z.unconsumed_tail + chunk)
                try:
                    decompressed = self.z.decompress(compressed, max_length)
                except zlib.error:
                    if not retry:
                        raise
                    self.z = zlib.decompressobj(-zlib.MAX_WBITS)
                    retry = False
                    decompressed = self.z.decompress(compressed, max_length)

                if not chunk:
                    self.flushed = self.z.flush()

            else:

                if not self.flushed:
                    return n

                decompressed = self.flushed[:max_length]
                self.flushed = self.flushed[max_length:]

            buf[n:n+len(decompressed)] = decompressed
            n += len(decompressed)
            max_length = len(buf) - n

        return n


class DeflateDecoder(ZDecoder):
    """ Decoding for "content-encoding: deflate" """


class GzipDecoder(ZDecoder):
    """ Decoding for "content-encoding: gzip" """

    def __init__(self, fp):
        ZDecoder.__init__(self, fp, zlib.decompressobj(16 + zlib.MAX_WBITS))



import re


ncr_replacements = {
    0x00: 0xFFFD,
    0x0D: 0x000D,
    0x80: 0x20AC,
    0x81: 0x0081,
    0x81: 0x0081,
    0x82: 0x201A,
    0x83: 0x0192,
    0x84: 0x201E,
    0x85: 0x2026,
    0x86: 0x2020,
    0x87: 0x2021,
    0x88: 0x02C6,
    0x89: 0x2030,
    0x8A: 0x0160,
    0x8B: 0x2039,
    0x8C: 0x0152,
    0x8D: 0x008D,
    0x8E: 0x017D,
    0x8F: 0x008F,
    0x90: 0x0090,
    0x91: 0x2018,
    0x92: 0x2019,
    0x93: 0x201C,
    0x94: 0x201D,
    0x95: 0x2022,
    0x96: 0x2013,
    0x97: 0x2014,
    0x98: 0x02DC,
    0x99: 0x2122,
    0x9A: 0x0161,
    0x9B: 0x203A,
    0x9C: 0x0153,
    0x9D: 0x009D,
    0x9E: 0x017E,
    0x9F: 0x0178,
}

LT = '<'
GT = '>'
AMP = '&'
SLASH = '/'
HASH = '#'

cdata_tags = ('script', 'style')

class InvalidNumCharRefReplacer(object):
    """ Replaces invalid numeric character references in HTML.

    In the wild you can find HTML containing numeric character
    references where the numbers are not valid Unicode characters.
    Most often the numbers are windows-1252 characters.
    So we replace this invalid characters as best we can.
    Note that the Python 3 html.parser.HTMLParser also does this
    and so does the html5lib HTML parser.
    """

    def __init__(self, fp):
        self.fp = fp
        self.clean = ''
        self.dirty = ''
        self.cdata_tag = None

    def read(self, size=None):
        while True:
            dirty = self.fp.read(size)
            eof = (size is None or not dirty)
            dirty = self.dirty + dirty
            clean, self.dirty, self.cdata_tag = clean_ncr(dirty,
                                                          eof,
                                                          self.cdata_tag)
            self.clean += clean
            if eof:
                assert not self.dirty
                data = self.clean
                self.clean = ''
                return data
            elif len(self.clean) >= size:
                data = self.clean[:size]
                self.clean = self.clean[size:]
                return data


def clean_ncr(dirty, eof, cdata_tag=None):
    parts = []
    # tokenize 'dirty' looking for start tags and numeric character references
    pos = 0
    # position of start of next part
    part_pos = 0

    assert not cdata_tag or cdata_tag == cdata_tag.lower().strip()

    end_token = {
        AMP: end_char_ref,
        LT: end_tag,
    }

    while True:

        if cdata_tag:
            # ignore char refs inside <script> or <style> tags.
            # just looking for </script> or </style>
            begin_token = begin_tag
        else:
            begin_token = begin_tag_or_char_ref

        begin = begin_token.search(dirty, pos)
        if not begin:
            # now we know there are not more tokens in this chunk
            pos = len(dirty)
            break

        token = end_token[begin.group()].search(dirty, begin.end())
        if not token:
            # We haven't seen the end of this token yet,
            # we may be called again, but with more data (if !eof).
            break

        if begin.group() == LT:
            if token.group(1) == SLASH and cdata_tag:
                tag = token.group(2).lower()
                if tag == cdata_tag:
                    cdata_tag = None
            else:
                tag = token.group(2).lower()
                if not cdata_tag and tag in cdata_tags:
                    cdata_tag = tag
        else:
            assert begin.group() == AMP
            assert not cdata_tag

            ncr = token.group(1)
            if ncr:
                if ncr.lower().startswith('x'):
                    code_point = int(ncr[1:], 16)
                else:
                    code_point = int(ncr, 10)

                if code_point in ncr_replacements:
                    parts.append(dirty[part_pos:begin.start()])
                    ncr = '&#x%0X' % ncr_replacements[code_point]
                    parts.append(ncr)
                    part_pos = token.end()

        pos = token.end()

    if eof:
        parts.append(dirty[part_pos:])
        pos = len(dirty)
    else:
        parts.append(dirty[:pos])

    return ''.join(parts), dirty[pos:], cdata_tag

#
# When scanning for the begining of tokens we look just one character
# at a time so that we don't have to worry about partial buffering.
begin_tag = re.compile(LT)
begin_tag_or_char_ref = re.compile('[<&]', re.I)
end_tag = re.compile('(/?)\s*([a-z]+)(?=[^a-z])', re.I)
end_char_ref = re.compile('[^#]|#(x[0-9A-F]+|[0-9]+)', re.I)


def replace_invalid_ncr(fp):
    return InvalidNumCharRefReplacer(fp)



"""Wrapper around multiprocessing.Pool to perform extraction"""

import logging
from six.moves import map
from functools import wraps
from contextlib import contextmanager
from multiprocessing import Pool


class MoreWork(Exception):
    """\
    Raised when more work needs to be added to the pool mid-run.

    An example of this would be a response containing a zip. We
    could raise this exception to send processing for files
    within the zip back to the pool for (multi-)processing.

    The 'work' parameter must:
        1. be pickleable
        2. give a sequence of (func, iterable) pairs suitable
           for using to extend the work list.
    """
    def __new__(cls, work):
        instance = Exception.__new__(cls)
        instance.work = work
        return instance


def do_func(f):

    @wraps(f)
    def do_func_wrapper(args):
        """\
        Manages exception propagation in work functions.

        We want to ensure that the parent process sees exceptions
        from child processes.

        For exceptions with specific meanings we just return them.
        For any other exception we also log it so we will see the
        traceback (if logging is configured correctly).
        """

        try:

            if isinstance(args, Exception):
                # yield_exc exception re-routing work-around
                # Coverage never sees this because it only
                # runs in child process, but the tests prove
                # it has happened.
                return args # pragma: no cover

            # Notice how we recieve "args" but call using "*args"
            return f(*args)

        except IOError as exc:
            return exc
        except SystemExit as exc:
            return exc
        except KeyboardInterrupt as exc:
            return exc
        except MoreWork as exc:
            return exc
        except BaseException as exc:
            logging.getLogger(__name__).exception('doing work')
            return exc

    return do_func_wrapper


@contextmanager
def close_and_shutdown(pool):
    try:
        yield pool
    except:
        pool.terminate()
        raise
    else:
        pool.close()
    finally:
        pool.join()


def yield_exc(iterable):
    """\
    part 1 of work-around for processes hanging if we get an exception during
    iteration.
    """
    try:
        for item in iterable:
            yield item
    except Exception as exc:
        yield exc


def do_in_pool(worklist, pool_size, initializer, initargs):
    with close_and_shutdown(Pool(pool_size, initializer, initargs)) as pool:
        while worklist:
            func, iterable = worklist.pop()
            for exc_or_none in pool.imap_unordered(func, yield_exc(iterable)):
                yield exc_or_none


# Special handling for single process "pool".
def do_in_this_process(work, initializer, initargs):
    if initializer is not None:
        initializer(*initargs)
    while work:
        func, iterable = work.pop()
        for exc in map(func, iterable):
            yield exc


def do(func, iterable, pool_size=None, initializer=None, initargs=()):
    """ Send work to the pool. """
    worklist = [(func, iterable)]

    if pool_size != 1:
        results = do_in_pool(worklist, pool_size, initializer, initargs)
    else:
        # This is especially useful for debugging
        results = do_in_this_process(worklist, initializer, initargs)

    for exc_or_none in results:
        if exc_or_none is None:
            # There is no exception - keep calm and carry on
            continue
        if isinstance(exc_or_none, MoreWork):
            worklist.extend(exc_or_none.work)
        else:
            assert isinstance(exc_or_none, BaseException)
            raise exc_or_none



""" HTMLStream converts a byte stream in to a unicode stream for parsing. """

from __future__ import unicode_literals, print_function
import codecs
from six.moves import map
from lxml.etree import XMLSyntaxError
from lxml.html import HTMLParser

CHUNK_SIZE = 1024
MAX_HEAD_CHUNKS = 50

#
# To process an HTML byte stream we need to know the character encoding.
# (often incorrectly called the charset.)
#
# Declarations of the encoding of an HTML document can be found:
#   * as a parameter in HTTP "Content-type" header
#   * in the "charset" attribute of an HTML <meta> tag
#   * in the "content" attribute of an HTML <meta> tag
#     with an "http-equiv" attribute of "content-type"
#
# Sometimes these will match, sometimes they will not.
# To find encodings in <meta> tags we need to pre-parse the document.


# stop pre-parsing if we reach any tags other than these
HEAD_TAGS = set([
    'html',
    'head',
    'title',
    'style',
    'base',
    'link',
    'meta',
    'script',
    'noscript',
])

# pre-parse this many bytes at a time
PRE_PARSE_CHUNK_SIZE = 4 * 1024

# stop pre-parsing if we have read this many bytes
MAX_PRE_PARSE_BYTES = 40 * 1024


# encodings where the (printable) characters are a superset
encoding_substitutions = {
    'gb2312': 'gbk',
    'iso8859-1': 'cp1252',
    'iso8859-9': 'cp1254',
}

#
# Take 10K random strings of 8 bytes each and
# see how many give you UnicodeDecode errors.
# Anything that isn't in the table gave 0.
ranking = {
    'cp1251': 3058,
    'iso8859-7': 9059,
    'cp1250': 14390,
    'cp1252': 14429,
    'cp1254': 19754,
    'iso8859-11': 22290,
    'tis-620': 24760,
    'cp1255': 52721,
    'shift_jis': 71873,
    'gbk': 74930,
    'big5': 90760,
    'euc_kr': 96158,
    'gb2312': 96714,
    'euc_jp': 96995,
    'utf-8': 99095,
    'ascii': 99627,
}


BOM_ENC = {
    codecs.BOM_UTF8: codecs.lookup('utf-8').name,
    codecs.BOM_UTF16_BE: codecs.lookup('utf-16-be').name,
    codecs.BOM_UTF16_LE: codecs.lookup('utf-16-le').name,
    codecs.BOM_UTF32_BE: codecs.lookup('utf-32-be').name,
    codecs.BOM_UTF32_LE: codecs.lookup('utf-32-le').name,
}


class HTMLStream(object):

    def __init__(self, response, filename=None):
        self.bom = None
        self.filename = filename
        self.encodings = []
        self.response = response
        self.declared_encodings = self.pre_parse()
        self.decoders = self.yield_decoders()
        self.next_encoding()

    def ranked_encodings(self):
        """ Declared encodings ranked by order of resistance to errors. """
        def lookup(encoding):
            try:
                name = codecs.lookup(encoding).name
                return encoding_substitutions.get(name, name)
            except LookupError:
                return None

        encodings = (encoding for label, encoding in self.declared_encodings)
        normalized = set(filter(None, map(lookup, encodings)))

        def key(enc):
            return ranking.get(enc, 0)

        # we want to try the most resilient encodings first
        return sorted(normalized, key=key, reverse=True)

    def yield_decoders(self):

        ranked_encodings = self.ranked_encodings()

        for encoding in ranked_encodings:
            info = codecs.lookup(encoding)
            decoder = info.incrementaldecoder()
            yield info.name, decoder

        if 'utf-8' not in ranked_encodings:
            info = codecs.lookup('utf-8')
            yield info.name, info.incrementaldecoder()

        # character set detection could go here
        if ranked_encodings:
            fallback = codecs.lookup(ranked_encodings[0])
        else:
            fallback = codecs.lookup('cp1252')

        # for our fallback we 'replace' errors
        yield fallback.name, fallback.incrementaldecoder('replace')

    def next_encoding(self):
        self.response.seek(0)
        self.decoded = u''
        self.encoding, self.decoder = next(self.decoders)
        self.strip_bom = self.bom

    def read(self, size=None):
        while True:
            raw_bytes = self.response.read(size)
            if not raw_bytes:
                # tell the decoder to flush
                self.decoded += self.decoder.decode(raw_bytes, True)
                break
            if self.strip_bom:
                self.strip_bom = False
                raw_bytes = raw_bytes[len(self.bom):]
            self.decoded += self.decoder.decode(raw_bytes)
            if size is None or len(self.decoded) >= size:
                break
        decoded = self.decoded[:size]
        self.decoded = self.decoded[len(decoded):]
        return decoded

    def pre_parse(self):

        http_content_type = self.response.headers.get('content-type', '')
        target = HTMLEncodings(http_content_type)
        # parser will fail on non-ascii unless we set it explicitly
        parser = HTMLParser(target=target, encoding='ISO-8859-1')
        total_bytes = 0

        self.response.seek(0)
        while target:
            chunk = self.response.read(PRE_PARSE_CHUNK_SIZE)
            if not chunk:
                try:
                    parser.close()
                except XMLSyntaxError:
                    pass
                break

            if self.bom is None:
                assert PRE_PARSE_CHUNK_SIZE >= 4
                self.bom = b''
                for i in range(4, 1, -1):
                    if chunk[:i] in BOM_ENC:
                        self.bom = chunk[:i]
                        target.encodings.append(('bom', BOM_ENC[self.bom]))
                        # the can only be one BOM - stop here
                        break

            parser.feed(chunk)
            total_bytes += len(chunk)
            if total_bytes >= MAX_PRE_PARSE_BYTES:
                break

        return target.encodings


class HTMLEncodings(object):

    def __init__(self, http_content_type):
        self.encodings = []
        self.parse_content_type(http_content_type, 'http-content-type')
        self.more = True

    def __bool__(self):
        return self.more

    def parse_content_type(self, content_type, label):
        # note: we have found at least one example with two "charset=" params
        #       where one was good and one was bad.
        #       HTTPMessage.get_content_charset() just takes the
        #       first, but actually we might as well try them all.
        for param in content_type.split(';')[1:]:
            key, _, value = param.partition('=')
            if key.strip().lower() == 'charset':
                self.encodings.append((label, value.strip()))

    def start(self, tag, attrib):
        if tag not in HEAD_TAGS:
            # we don't want any more chunks
            self.more = False
        elif tag == 'meta':
            if 'charset' in attrib:
                self.encodings.append(('meta-charset', attrib['charset']))
            elif attrib.get('http-equiv', '').lower() == 'content-type':
                content_type = attrib.get('content', '')
                self.parse_content_type(content_type, 'meta-content-type')

    def close(self):
        pass



""" Extractors for URLs from 
`/robots.txt <http://en.wikipedia.org/wiki/Robots_exclusion_standard#Sitemap>`_
and `sitemaps <http://www.sitemaps.org/protocol.html>`_.
"""


from __future__ import unicode_literals, absolute_import, print_function
import logging
import wex.py2compat ; assert wex.py2compat
from lxml.etree import iterparse, XMLSyntaxError
from codecs import getreader
from six.moves.urllib_parse import urljoin
from wex.extractor import chained
from wex.http import decode
from wex.url import URL

log = logging.getLogger(__name__)


def urls_from_robots_txt(response):
    """ Yields sitemap URLs from "/robots.txt" """

    url = URL(response.request_url or response.url or '')
    if url.parsed.path != '/robots.txt':
        return

    charset = response.headers.get_content_charset()
    lines = getreader(charset or 'ISO-8859-1')(response)

    for line in lines:

        content, _, comment = line.partition('#')
        field, _, value = content.partition(':')
        if field.strip().lower() != 'sitemap':
            continue

        # we shouldn't need to urljoin but we do just in case
        joined = URL(urljoin(response.url, value.strip()))
        # set sitemap=True in fragment to help downstream processing
        yield "url", joined.update_fragment_dict(sitemap=True)


def urls_from_urlset_or_sitemapindex(response):
    """ Yields URLs from ``<urlset>`` or ``<sitemapindex>`` elements as per 
        `sitemaps.org <http://www.sitemaps.org/protocol.html>`_.
    """

    sitemap = URL(response.url).fragment_dict.get('sitemap')
    content_subtypes = response.headers.get_content_subtype().split('+')
    if not sitemap and 'xml' not in content_subtypes:
        return

    root = None
    try:
        for _, elem in iterparse(decode(response)):

            if root is None:
                root = elem.getroottree().getroot()
                if not (root.tag.endswith('}sitemapindex') or
                        root.tag.endswith('}urlset')):
                    # root element has wrong tag - give up
                    break

            if elem.tag.endswith('}loc') and elem.text is not None:
                text = elem.text.strip()
                if text:
                    # http://www.sitemaps.org/protocol.html#locdef
                    url = URL(urljoin(response.url, text))
                    if elem.getparent().tag.endswith('}sitemap'):
                        # set sitemap=True to help downstream processing
                        url = url.update_fragment_dict(sitemap=True)
                    yield "url", url

            if elem.getparent() is root:
                # release memory for previous elements
                while elem.getprevious() is not None:
                    del root[0]

    except XMLSyntaxError:
        log.debug("XMLSyntaxError in '%s' (%d)", response.url, response.code)

#: Extractor that combines :func:`.urls_from_robots_txt` and
#: :func:`.urls_from_urlset_or_sitemapindex`.
urls_from_sitemaps = chained(urls_from_robots_txt,
                             urls_from_urlset_or_sitemapindex)



import os
import logging
import json
from six import binary_type
from six.moves.urllib_parse import urlparse
from threading import Timer
from subprocess import Popen, PIPE
from pkg_resources import resource_filename

TIMEOUT = 60.0

script = os.path.abspath(resource_filename(__name__, 'js/phantom.js'))
cmd = ['phantomjs', '--ssl-protocol=any', script]
# see http://phantomjs.org/api/webpage/property/settings.html
default_settings = {'loadImages': False}


def request_using_phantomjs(url, method, session=None, **kw):

    phantomjs = Popen(cmd, stdin=PIPE, stdout=PIPE)

    proxies = kw.get('proxies', None)
    if proxies:
        # only 'http' proxies for now
        proxy_url = proxies['http']
        parsed = urlparse(proxy_url)
        proxy = {
            'type': 'http',
            'hostname': parsed.hostname,
            'port': parsed.port or 80,
            'username': parsed.username,
            'password': parsed.password,
        }
    else:
        proxy = None

    def terminate_phantomjs():
        if phantomjs.poll() is not None:
            return
        phantomjs.terminate()
        logging.getLogger(__name__).warning("phantomjs terminated")

    timeout_timer = Timer(TIMEOUT, terminate_phantomjs)
    # we don't want the main thread to wait for us
    timeout_timer.daemon = True

    settings = dict(default_settings)
    settings.update(method.args.get('settings', {}))
    requires = []
    for require in method.args.get('requires', []):
        if isinstance(require, (tuple, list)):
            stem, _ = os.path.splitext(resource_filename(*require))
            requires.append(stem)
        else:
            requires.append(require)

    request = {
        "url": url,
        "requires": requires,
        "settings": settings,
        "loglevel": logging.getLogger(__name__).getEffectiveLevel(),
        "context": kw.get("context", {}),
        "proxy": proxy,
        "args": method.args,
    }
    dumped = json.dumps(request)
    if not isinstance(dumped, binary_type):
        # Python 3 json.dumps produces unicode, but stdin needs binary
        dumped = dumped.encode('utf-8')
    phantomjs.stdin.write(dumped)
    phantomjs.stdin.close()
    timeout_timer.start()
    yield phantomjs.stdout
    timeout_timer.cancel()



from __future__ import unicode_literals
from wex.iterable import flatten
from operator import methodcaller

strip = methodcaller('strip')

def partition(separator, **kw):
    """ Returns a function that yields tuples created by partitioning
        text using `separator`.
    """
    normalize_head = kw.pop('normalize_head', strip)
    normalize_tail = kw.pop('normalize_tail', strip)
    def _partition(obj):
        for s in flatten(obj):
            head, sep, tail = s.partition(separator)
            if normalize_head:
                head = normalize_head(head)
            if normalize_tail:
                tail = normalize_tail(tail)
            if sep:
                yield (head, tail)
    return _partition



"""
Functions creating readable objects.

These objects are readable because they have `readline` and `read` methods.
They are used to to create `Response` objects using `Response.from_readable`.
"""

from __future__ import absolute_import, unicode_literals, print_function
import os
import errno
import sys
import tarfile
from io import FileIO
from threading import local
from functools import partial as partial_
from contextlib import closing
from six import PY2
from . import py2compat  # flake8: noqa
from .url import URL


EXT_WEXIN = '.wexin'
LF = b'\n'


class partial(partial_):
    def __repr__(self):
        """ Customized __repr__ for use with `Open` class. """
        return '%r, %r' % (self.func, self.args)


class Open(object):
    """ Lazily open a readable, when asked for read/readline.

    The reason we want to lazily open the readable is because
    that makes this object pickleable in its pre-opened state
    and that allows us to send the readable into a
    `multiprocessing.Pool`.
    """

    def __init__(self, open):
        self.open = open

    def __repr__(self):
        return 'Open(%r)' % self.open

    def __getattr__(self, name):
        if name not in ('readline', 'read', 'close', 'name'):
            raise AttributeError
        fp = self.open()

        self.__dict__['read'] = fp.read
        self.__dict__['readline'] = fp.readline
        self.__dict__['close'] = fp.close
        if hasattr(fp, 'name'):
            self.__dict__['name'] = fp.name

        assert name in self.__dict__
        return self.__dict__[name]

_open_tarfile = local()
_open_tarfile.tarfile = None


def tarfile_open(path):
    tarfileobj = getattr(_open_tarfile, 'tarfile', None)
    if tarfileobj is not None:
        if tarfileobj.name == path and not tarfileobj.closed:
            # same tarfile - no need to re-open
            return tarfileobj
        # there shouldn't be any problem closing this
        # because the fact that we have been asked to
        # open another tarfile will always mean that this
        # process has moved on to a different file.
        tarfileobj.close()

    _open_tarfile.tarfile = tarfile.open(path, 'r')
    return _open_tarfile.tarfile


def tarfile_tarinfo_open(path, tarinfo):
    tf = tarfile_open(path)
    return tf.extractfile(tarinfo)


def readables_from_paths(paths, save_dir=None):
    """ Yield readables from a sequence of paths """

    for path in paths:
        if path.strip() == '-':
            stdin = sys.stdin if PY2 else sys.stdin.buffer
            tar = tarfile.open(mode='r|*', fileobj=stdin)
            for member in tar:
                yield tar.extractfile(member)
        elif not os.path.exists(path) and URL(path).parsed.scheme:
            url = URL(path)
            readables = url.get()
            if save_dir:
                readables = save_readables(url, save_dir, readables)
            for readable in readables:
                yield readable
        else:
            for readable in readables_from_file_path(path):
                yield readable


def save_readables(url, save_dir, readables):
    url_dir = url.mkdirs(save_dir)
    for i, readable in enumerate(readables):
        basename = '{}{}'.format(i, EXT_WEXIN)
        path = os.path.join(url_dir, basename)
        fp = FileIO(path, 'w')
        readable = TeeReadable(readable, fp)
        with closing(readable):
            yield readable


def readables_from_file_path(path):
    """ Yield readables from a file system path """

    numdirs = 0
    for dirpath, dirnames, filenames in os.walk(path):
        numdirs += 1
        # Don't walk into "hidden" directories
        dirnames[:] = [n for n in dirnames if not n.startswith('.')]
        for filename in filenames:
            if filename.lower().endswith(EXT_WEXIN):
                filepath = os.path.join(dirpath, filename)
                yield Open(partial(FileIO, filepath))

    if numdirs < 1:
        if path.lower().endswith('EXT_WEXIN'):
            yield Open(partial(FileIO, path))
        else:
            try:
                tf = tarfile_open(path)
                for readable in readables_from_tarfile(tf):
                    yield readable
            except IOError as exc:
                if exc.errno != errno.ENOENT:
                    raise
                # we want to defer the ENOENT error to later
                # so we do that by yielding an Open(...)
                yield Open(partial(FileIO, path))
            except tarfile.ReadError:
                yield Open(partial(FileIO, path))


def readables_from_tarfile(tf):
    for ti in tf:
        if ti.name.endswith(EXT_WEXIN):
            yield Open(partial(tarfile_tarinfo_open, tf.name, ti))


class TeeReadable(object):
    """ Readable that writes out to a tee file. """

    def __init__(self, readable, tee):
        self.readable = readable
        self.tee = tee

    @property
    def name(self):
        return self.tee.name

    def read(self, size):
        buf = self.readable.read(size)
        if buf:
            self.tee.write(buf)
        return buf

    def readline(self, *args):
        buf = self.readable.readline(*args)
        if buf:
            self.tee.write(buf)
        return buf

    def close(self):
        if not getattr(self.readable, 'closed', False):
            # read remainder so tee gets everything
            chunk_size = 2**16
            while True:
                chunk = self.read(chunk_size)
                if not chunk:
                    break
        self.readable.close()
        self.tee.close()


class ChainedReadable(object):
    """ Readable that combines the contents of multiple filelike together """

    EMPTY = b''

    def __init__(self, *files):
        self.files = files
        self.n = 0

    def read(self, size):
        buf = self.EMPTY
        while len(buf) < size and self.n < len(self.files):
            chunk = self.files[self.n].read(size - len(buf))
            if not chunk:
                self.n += 1
            buf += chunk
        return buf

    def readline(self, *args):
        if self.n >= len(self.files):
            return self.EMPTY
        buf = self.files[self.n].readline(*args)
        while LF not in buf and (not args or len(buf) < args[0]):
            self.n += 1
            if self.n >= len(self.files):
                break
            # OK so this can end up reading a line > limit but it isn't
            # going to really matter so we just ignore the possibiity.
            # We don't want to have to manage remainders unless we have to.
            buf += self.files[self.n].readline(*args)
        return buf

    def close(self):
        for fileobj in self.files:
            fileobj.close()



from setuptools import setup
from wex import __version__

setup(
    name='Wextracto',
    version=__version__,
    description='Web Data Extraction Library Written in Python',
    long_description='\n\n'.join([
        open('README.rst').read(),
        open('HISTORY.rst').read(),
    ]),
    author='Giles Brown',
    author_email='gsbrown@ebay.com',
    url='https://github.com/eBay/wextracto',
    download_url='https://github.com/eBay/wextracto/tarball/' + __version__,
    license='BSD',
    include_package_data=True,
    packages=['wex'],
    package_data={
        '': ['LICENSE.txt', 'NOTICES.txt'],
        'wex': ['logging.conf', 'phantomjs.js', 'js/*.js'],
    },
    zip_safe=False,
    install_requires=[
        'six',
        'requests',
        'lxml>=3',
        'cssselect',
        'publicsuffix>=1.1',
    ],
    classifiers=(
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
    ),
    entry_points="""
        [console_scripts]
        wex = wex.command:main

        [pytest11]
        wex = wex.pytestplugin

        [wex.method.http]
        get = wex.http:request
        post = wex.http:request
        phantomjs = wex.phantomjs:request_using_phantomjs
        form = wex.form:submit_form

        [wex.method.https]
        get = wex.http:request
        post = wex.http:request
        phantomjs = wex.phantomjs:request_using_phantomjs
        form = wex.form:submit_form

        [wex.method.ftp]
        get = wex.ftp:get
    """
)
